"""
Backward compatibility module for qhub.qiskit imports.
This module provides access to the quantum providers for existing code.
"""

from qhub.quantum.sdk.qiskit.backend import HubQiskitBackend
from qhub.quantum.sdk.qiskit.job import HubQiskitJob
# Import from the new quantum structure for backward compatibility
from qhub.quantum.sdk.qiskit.provider import HubQiskitProvider

# Re-export for backward compatibility
__all__ = ['HubQiskitProvider', 'HubQiskitBackend', 'HubQiskitJob']
